from collections import defaultdict

import pytest
import tribune.sheet_import.parsers as p
from .utils import assert_import_errors


def assert_really_equal(left, right):
    assert type(left) == type(right)
    assert left == right


def test_chain():
    assert p.chain(lambda _: 'Python')(0.99) == 'Python'
    assert p.chain(lambda x: x + 1, lambda x: x - 1)(0) == 0
    assert p.chain(lambda x: x * 2, lambda x: x - 1)(1) == 1
    assert p.chain(lambda x: x - 1, lambda x: x * 2)(1) == 0


def test_parse_text():
    assert p.parse_text('') == u''
    assert p.parse_text('  ') == u''
    assert p.parse_text(' A ') == u'A'


def test_parse_yes_no():
    assert p.parse_yes_no('yes') is True
    assert p.parse_yes_no('YES') is True
    assert p.parse_yes_no('  YES  ') is True
    assert p.parse_yes_no('no') is False
    assert p.parse_yes_no('NO') is False
    assert p.parse_yes_no('  NO  ') is False

    for case in {'yen', 'YEN', ''}:
        assert_import_errors({'must be "yes" or "no"'}, lambda: p.parse_yes_no(case))


def test_parse_int():
    assert_really_equal(p.parse_int('1'), 1)
    assert_really_equal(p.parse_int('-1'), -1)
    assert_really_equal(p.parse_int('123'), 123)
    assert_really_equal(p.parse_int('-123'), -123)
    assert_really_equal(p.parse_int('39.000'), 39)
    assert_really_equal(p.parse_int('-39.000'), -39)
    assert_really_equal(p.parse_int('1e+3'), 1000)
    assert_really_equal(p.parse_int(1), 1)

    for case in {'', 'a', '1ob', 'ob1', '3.2', 'one'}:
        assert_import_errors({'must be an integer'}, lambda: p.parse_int(case))


def test_parse_int_as_text():
    assert_really_equal(p.parse_int_as_text('1'), u'1')


def test_parse_number():
    assert_really_equal(p.parse_number('1'), 1.0)
    assert_really_equal(p.parse_number('-1'), -1.0)
    assert_really_equal(p.parse_number('3.9'), 3.9)
    assert_really_equal(p.parse_number('-3.9'), -3.9)
    assert_really_equal(p.parse_number('1e-3'), 0.001)
    assert_really_equal(p.parse_number(1.0), 1.0)

    for case in {'', 'a1', '1a', 'one'}:
        assert_import_errors({'must be a number'}, lambda: p.parse_number(case))


def test_parse_int_or_text():
    assert_really_equal(p.parse_int_or_text(''), u'')
    assert_really_equal(p.parse_int_or_text('   '), u'')
    assert_really_equal(p.parse_int_or_text('1'), u'1')
    assert_really_equal(p.parse_int_or_text('1.0'), u'1')
    assert_really_equal(p.parse_int_or_text(' 937'), u'937')
    assert_really_equal(p.parse_int_or_text('1e3'), u'1000')
    assert_really_equal(p.parse_int_or_text('a b c'), u'a b c')
    assert_really_equal(p.parse_int_or_text('  a b c  '), u'a b c')


def test_default_to():
    assert p.default_to('1')('') == '1'
    assert p.default_to('1')('  ') == '1'
    assert p.default_to('1')('1') == '1'
    assert p.default_to('1')('2') == '2'
    assert p.default_to(1)(None) == 1


def test_nullable():
    assert p.nullable(lambda _: 10)('') is None
    assert p.nullable(lambda _: 10)(None) is None
    assert p.nullable(lambda _: 10)('  ') is None
    assert p.nullable(lambda _: 10)('a') == 10
    assert p.nullable(lambda _: 10)(False) == 10
    assert p.nullable(lambda _: 10)(0) == 10


def test_validate_satisfies():
    def always(_):
        return True

    def never(_):
        return False

    assert p.validate_satisfies(always, '')('some-value') == 'some-value'
    assert p.validate_satisfies(always, '')(9) == 9

    assert_import_errors({'nope'}, lambda: p.validate_satisfies(never, 'nope')('some-value'))
    assert_import_errors({'nope'}, lambda: p.validate_satisfies(never, 'nope')(9))

    is_two = p.validate_satisfies(lambda x: x == 2, 'must be 2')
    assert is_two(2) == 2
    assert_import_errors({'must be 2'}, lambda: is_two(3))


def test_validate_not_empty():
    assert p.validate_not_empty('something') == 'something'
    assert_import_errors({'must not be empty'}, lambda: p.validate_not_empty(''))
    assert_import_errors({'must not be empty'}, lambda: p.validate_not_empty('  '))


def test_validate_max_length():
    assert p.validate_max_length(1)('a') == 'a'
    assert p.validate_max_length(2)('a') == 'a'
    assert p.validate_max_length(200)('a') == 'a'
    assert p.validate_max_length(200)('a'*200) == 'a'*200
    assert_import_errors({'must be no more than 1 characters long'},
                         lambda: p.validate_max_length(1)('ab'))
    assert_import_errors({'must be no more than 200 characters long'},
                         lambda: p.validate_max_length(200)('a'*201))


def test_validate_min():
    assert p.validate_min(-30)(-29) == -29
    assert p.validate_min(0)(0) == 0
    assert p.validate_min(1)(1) == 1
    assert p.validate_min(1)(2) == 2
    assert p.validate_min(1)(3.9) == 3.9
    assert p.validate_min(3.8)(3.9) == 3.9
    assert p.validate_min(float('-inf'))(1) == 1
    assert p.validate_min(100)(100) == 100

    assert_import_errors({'number must be no less than 12'},
                         lambda: p.validate_min(12)(11))
    assert_import_errors({'number must be no less than -30'},
                         lambda: p.validate_min(-30)(-31))
    assert_import_errors({'number must be no less than 3.9'},
                         lambda: p.validate_min(3.9)(3.8))


def test_validate_max():
    assert p.validate_max(-30)(-31) == -31
    assert p.validate_max(1)(1) == 1
    assert p.validate_max(3.9)(3.9) == 3.9
    assert p.validate_max(float('inf'))(1) == 1
    assert p.validate_max(100)(100) == 100

    assert_import_errors({'number must be no greater than 12'},
                         lambda: p.validate_max(12)(13))
    assert_import_errors({'number must be no greater than -30'},
                         lambda: p.validate_max(-30)(-29))
    assert_import_errors({'number must be no greater than 3.9'},
                         lambda: p.validate_max(3.9)(4))


def test_validate_range():
    assert p.validate_range(-30, 30)(0) == 0

    assert_import_errors({'number must be no greater than 30'},
                         lambda: p.validate_range(-30, 30)(31))
    assert_import_errors({'number must be no less than -30'},
                         lambda: p.validate_range(-30, 30)(-31))


def test_validate_unique():
    assert p.validate_unique({})(1) == 1
    assert p.validate_unique({})('abc') == 'abc'
    assert p.validate_unique({})(' abc ') == ' abc '
    assert p.validate_unique({1})(2) == 2
    assert p.validate_unique({2})(1) == 1
    assert p.validate_unique({1, 2})(3) == 3
    assert p.validate_unique(set(range(100)))(101) == 101

    fail_cases = [({1}, 1), ({1, 2}, 2), (set(range(100)), 99)]

    for others, value in fail_cases:
        with pytest.raises(p.SpreadsheetImportError):
            p.validate_unique(others)(value)


def test_validate_one_of():
    assert p.validate_one_of({1})(1) == 1
    assert p.validate_one_of({2})(2) == 2
    assert p.validate_one_of({1, 2})(1) == 1
    assert p.validate_one_of({1, 2})(2) == 2

    fail_cases = [({1}, 2), ({1, 2}, 3), (set(range(100)), 101)]

    for others, value in fail_cases:
        with pytest.raises(p.SpreadsheetImportError):
            p.validate_one_of(others)(value)

    assert_import_errors({u'must be a valid thing'}, lambda: p.validate_one_of({1})(2))
    assert_import_errors({u'must be a valid chicken'},
                         lambda: p.validate_one_of({1}, thing='chicken')(2))
    assert_import_errors(
        {u'must be a valid chicken: 1, 2, 3'},
        lambda: p.validate_one_of([1, 2, 3], thing='chicken', show_choices_in_error=True)(4)
    )


def test_parse_list():
    assert p.parse_list(delim=' ')('') == [u'']
    assert p.parse_list(delim=' ')('a') == [u'a']
    assert p.parse_list(delim=' ')('  a   ') == [u'a']
    assert p.parse_list(delim=',')('  a   ') == [u'a']
    assert p.parse_list(delim=' ')('a b') == [u'a', u'b']
    assert p.parse_list(delim=',')('a b') == [u'a b']
    assert p.parse_list(delim=',')('a, b, c') == [u'a', u' b', u' c']
    assert p.parse_list(p.parse_text, delim=',')('a, b, c') == [u'a', u'b', u'c']
    assert p.parse_list(lambda x: x.upper(), delim=',')('a,b,c') == [u'A', u'B', u'C']


def test_parse_lookup():
    assert p.parse_lookup({'a': 1})('a') == 1
    assert p.parse_lookup({'a': 1, 'b': 2})('b') == 2
    assert p.parse_lookup(defaultdict(lambda: None, {'a': 1}))('a') == 1
    assert p.parse_lookup(defaultdict(lambda: None, {'a': 1}))('b') is None


def test_filter_non_empty():
    assert p.filter_non_empty([]) == []
    assert p.filter_non_empty([1]) == [1]
    assert p.filter_non_empty([0, 1]) == [0, 1]
    assert p.filter_non_empty([False, True]) == [False, True]
    assert p.filter_non_empty([None]) == []
    assert p.filter_non_empty(['']) == []
    assert p.filter_non_empty(['', 1, None]) == [1]
    assert p.filter_non_empty((1, '', None, 2)) == [1, 2]



import tribune.sheet_import as si
import tribune.sheet_import.parsers as p
from .utils import assert_import_errors


def test_normalize_text():
    assert p.normalize_text('') == ''
    assert p.normalize_text('a') == 'a'
    assert p.normalize_text('a b') == 'a b'
    assert p.normalize_text('   ') == ''
    assert p.normalize_text(' a b') == 'a b'
    assert p.normalize_text('a b ') == 'a b'
    assert p.normalize_text(' a b ') == 'a b'
    assert p.normalize_text('A b') == 'a b'
    assert p.normalize_text(' A B ') == 'a b'


def parse_records_with(importer):
    return lambda data: list(
        importer.get_sheet_records(
            si.XlrdSheetData.from_nested_iters(data)))


class TestSheetImport(object):
    def test_sheet_import_no_fields(self):
        parse_records = parse_records_with(si.SheetImporter([]))

        assert parse_records([]) == []
        assert parse_records([[]]) == []
        assert parse_records([['data']]) == []

    def test_sheet_import_simple(self):
        parse_records = parse_records_with(si.SheetImporter([
            si.Field('Name', 'name', p.parse_text),
        ]))

        assert parse_records([['Name']]) == []  # no data
        assert parse_records([['name']]) == []  # different header format
        assert parse_records([[' name ']]) == []  # different header format
        assert parse_records([['Name'], ['Ed']]) == [{'name': 'Ed'}]  # one row, one column
        assert parse_records([['Name'], ['Ed', '']]) == [{'name': 'Ed'}]  # one row, extra column
        assert parse_records([['Name'], ['Ed'], ['Joe']]) \
            == [{'name': 'Ed'}, {'name': 'Joe'}]  # two rows

        # invalid header
        invalid_headers = (
            [],                     # No data
            [['Some other name']],  # Wrong header name
        )
        for case in invalid_headers:
            assert_import_errors({'Expected "Name" in header cell A1.'},
                                 lambda: parse_records(case))

    def test_sheet_import_data_validation(self):
        parse_records = parse_records_with(si.SheetImporter([
            si.Field('Name', 'name', p.chain(p.parse_text, p.validate_not_empty)),
            si.Field('Age',  'age',  p.parse_int),
        ]))

        assert parse_records([['Name', 'Age']]) == []
        assert parse_records([['Name', 'Age'], ['Ed', '27']]) == [{'name': 'Ed', 'age': 27}]

        ed_n_joe = [
            {'name': 'Ed',  'age': 27},
            {'name': 'Joe', 'age': 32},
        ]

        # Test with and without extra columns or formatting
        assert parse_records([['Name', 'Age'], ['Ed', '27'], ['Joe', '32']]) == ed_n_joe
        assert parse_records([['Name', 'Age'], ['Ed', '27.0'], ['Joe', '32.00']]) == ed_n_joe
        assert parse_records([['Name', 'Age'], ['Ed', ' 27 '], ['Joe', ' 32.0 ']]) == ed_n_joe
        assert parse_records([['Name', 'Age'], ['Ed', '27', '2'], ['Joe', '32', '3']]) == ed_n_joe

        # invalid header and/or data
        assert_import_errors(
            {'Expected "Age" in header cell B1.'},
            lambda: parse_records([['Name', 'SSN']])
        )
        assert_import_errors(
            {'Expected "Age" in header cell B1.'},  # header errors beat data errors
            lambda: parse_records([['Name', 'SSN'], ['', '2']])
        )
        assert_import_errors(
            {'Unexpected value in cell A2: must not be empty',
             'Unexpected value in cell B2: must be an integer'},
            lambda: parse_records([['Name', 'Age'], ['', '3.9']])
        )
        assert_import_errors(
            {'Unexpected value in cell A3: must not be empty',
             'Unexpected value in cell B3: must be an integer'},
            lambda: parse_records([['Name', 'Age'], ['Ed', '27'], ['', '3.9']])
        )

    def test_non_string_header(self):
        parse_records = parse_records_with(si.SheetImporter([
            si.Field('Age', 'age', p.parse_int)
        ]))
        assert_import_errors(
            {'Expected "Age" in header cell A1.'},
            lambda: parse_records([[29]])
        )

    def test_sheet_import_with_modify_record(self):
        class TestImport(si.SheetImporter):
            fields = (
                si.Field('Name', 'name', p.chain(p.parse_text, p.validate_not_empty)),
            )

            def modify_record(self, record):
                if record['name'] == 'Ed':  # Filter out Eds
                    return None

                record['age'] = 27
                return record

        parse_records = parse_records_with(TestImport())

        assert parse_records([['Name']]) == []
        assert parse_records([['Name'], ['Ed']]) == []  # Ed is filtered out

        joe = [
            {'name': 'Joe', 'age': 27},  # age is always reset to 27
        ]

        assert parse_records([['Name'], ['Ed', '27'], ['Joe', '32']]) == joe
        assert parse_records([['Name'], ['Joe', '32.00'], ['Ed', '27.0']]) == joe

        # invalid header and/or data
        assert_import_errors(
            {'Expected "Name" in header cell A1.'},
            lambda: parse_records([['Age', 'Name']])
        )
        assert_import_errors(
            {'Unexpected value in cell A2: must not be empty'},
            lambda: parse_records([['Name'], [''], ['data']])
        )

    def test_sheet_import_data_start_row(self):
        parse_records = parse_records_with(si.SheetImporter([
            si.Field('Is Cool', 'is_cool', p.parse_yes_no)
        ], data_start_row=3))

        parse_records([
            [''],
            [''],
            ['Is Cool'],
            ['yes'],
            ['no'],
        ]) == [{'is_cool': True}, {'is_cool': False}]

        assert_import_errors(
            {'Expected "Is Cool" in header cell A3.'},
            lambda: parse_records([['Is Cool']])
        )

    def test_sheet_import_as_report_sheet(self):
        importer = si.SheetImporter((
            si.Field('Name', 'name', p.chain(p.parse_text, p.validate_not_empty)),
            si.Field('Job',  'job',  p.parse_text),
        ))

        TestReportSheet = importer.as_report_sheet()
        assert TestReportSheet.pre_data_rows == importer.data_start_row
        # TODO More tests here.



import copy
from decimal import Decimal
import sys
from types import FunctionType

from blazeutils.spreadsheets import WriterX
import six
from xlsxwriter.format import Format

from .utils import (
    column_letter,
    deepcopy_tuple_except
)

# we do not want to make SA a dependency, just want to work with it when it is used
try:
    from sqlalchemy.orm.attributes import QueryableAttribute
except ImportError:
    class QueryableAttribute(object):
        pass


__all__ = [
    'SheetUnit',
    'SheetColumn',
    'SheetPortraitColumn',
    'SheetSection',
    'BlankColumn',
    'LabeledColumn',
    'PortraitRow',
    'TotaledMixin',
    'ReportSheet',
    'ReportPortraitSheet',
]


def _fetch_val(record, key):
    if key is None:
        return ''
    elif isinstance(key, str) and hasattr(record, key):
        return getattr(record, key)
    elif isinstance(key, str) and (isinstance(record, dict) and key in record):
        return record[key]
    elif isinstance(key, tuple):
        # magic way to combine data fields in a cell. tuple is expected as
        #   (SA-col-1/string/int, SA-col-2/string/int, operator)
        val = key[2](
            _fetch_val(record, key[0]),
            _fetch_val(record, key[1])
        )
    elif not isinstance(key, (str, int, float, Decimal)):
        # must be a SA-col, find the key and hit the record
        datakey = getattr(key, 'key', getattr(key, 'name', None))
        val = getattr(record, datakey)
    else:
        # if none of the above, put the key in the cell
        val = key
    return val


class ProgrammingError(Exception):
    pass


class _SheetDeclarativeMeta(type):
    """
        Metaclass to define/create units on a sheet section
        Borrowed in part from WebGrid's Grid setup
    """

    def __new__(cls, name, bases, class_dict):
        class_units = []

        # add columns from base classes
        for base in bases:
            base_units = getattr(base, '__cls_units__', ())
            class_units.extend(base_units)
        class_units.extend(class_dict.get('__cls_units__', ()))
        class_dict['__cls_units__'] = class_units

        return super(_SheetDeclarativeMeta, cls).__new__(cls, name, bases, class_dict)


class SheetUnit(object):
    # Basic sheet object, can be a row or a column

    def __new__(cls, *args, **kwargs):
        col_inst = super(SheetUnit, cls).__new__(cls)
        if '_dont_assign' not in kwargs:
            col_inst._assign_to_section()
        return col_inst

    def new_instance(self, sheet):
        unit = copy.deepcopy(self)
        unit.sheet = sheet
        return unit

    def _assign_to_section(self):
        section_locals = sys._getframe(2).f_locals
        section_cls_units = section_locals.setdefault('__cls_units__', [])
        section_cls_units.append(self)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        # setting the id prevents self-references from deep copying
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            # SQLAlchemy attributes need special care for copying, don't deep copy these
            # Functions (FunctionType) also get the original object passed, as the copy lib does
            #   the same. This handling will avoid problems with decorated functions

            deepcopy_exceptions = (QueryableAttribute, FunctionType)

            if isinstance(v, deepcopy_exceptions):
                # we have a value that we don't want to deepcopy, just assign it
                setattr(result, k, v)

            elif isinstance(v, tuple):
                # tribune supports tuple expressions (see `_fetch_val`), which can
                # be nested arbitrarily deep. We want to deep-copy all of the nested
                # tuple, except for any QueryableAttributes
                setattr(result, k, deepcopy_tuple_except(v, deepcopy_exceptions))

            else:
                # fall back to default copy operation
                setattr(result, k, copy.deepcopy(v, memo))

        return result


class SheetColumn(SheetUnit):
    """
        Basic column for a spreadsheet report. Holds the SQLAlchemy
        column and has methods for rendering various kinds of rows
    """
    def new_instance(self, sheet):
        column = SheetUnit.new_instance(self, sheet)
        column._construct_header_data(self._init_header)

        return column

    def __init__(self, key=None, sheet=None, write_header_func=None, write_data_func=None,
                 write_total_func=None, **kwargs):
        """
            header_# in kwargs used to override default (blank) heading values
            write_*_func can override the class methods
        """
        # key can be one of many types: str, Decimal, int, float, SA column,
        #   or a tuple of the above (with an operator)
        self.expr = None
        if key is not None and not isinstance(key, (str, tuple, Decimal, int, float)):
            self.expr = col = key
            # use column.key, column.name, or None in that order
            key = getattr(col, 'key', getattr(col, 'name', None))

        self.key = key
        self.xls_width = kwargs.get('xls_width', getattr(self, 'xls_width', None))
        self.xls_computed_width = 0
        self.sheet = sheet
        self.write_header = write_header_func or self.write_header
        self.write_data = write_data_func or self.write_data
        self.write_total = write_total_func or self.write_total

        # look for header values in kwargs, construct a header dict
        self._init_header = dict()
        for k, v in six.iteritems(kwargs):
            if k.startswith('header_'):
                try:
                    self._init_header[int(k[7:])] = v
                except ValueError:
                    pass

    def _construct_header_data(self, init_header):
        d = ['' for i in range(self.sheet.pre_data_rows)]
        for k, v in six.iteritems(init_header):
            try:
                d[k] = v
            except IndexError:
                raise ProgrammingError('not enough pre-data rows on sheet')
        self.header_data = d

    def xls_width_calc(self, value):
        if self.xls_width:
            return self.xls_width
        if isinstance(value, str) and len(value) and value[0] == '=':
            # formulas start with an equals
            return 0
        if isinstance(value, str):
            return len(value)
        return len(str(value))

    def register_col_width(self, value):
        # keep track of the width of column contents
        if value is None:
            return
        self.xls_computed_width = max(
            self.xls_computed_width,
            self.xls_width_calc(value)
        )

    def adjust_col_width(self):
        # set column width according to computed width by register_col_width
        if hasattr(self, 'col_idx'):
            final_width = max(min(self.xls_computed_width, 150), 5)

            # width calculation is 1/256th of width of zero character, using the
            # first font that occurs in the Excel file
            # the following calculation seems to get it done alright
            self.sheet.ws.set_column(self.col_idx, self.col_idx, int(final_width + 3))

    def fetch_val(self, record, key):
        # wrapper for _fetch_val to make an instance override easier
        return _fetch_val(record, key)

    def extract_data(self, record):
        return self.fetch_val(record, self.key)

    def format_data(self, value=None):
        return None

    def format_header(self, header_row):
        # called for each pre_data_row row. header_row is zero-based.
        return None

    def format_total(self):
        return None

    def write_header(self):
        # set the column index to be used for adjustment of width
        # and write the stored header data from init
        self.col_idx = self.sheet.colnum
        val = self.header_data[self.sheet.rownum]
        self.register_col_width(val)
        self.sheet.awrite(val, self.format_header(self.sheet.rownum))

    def write_data(self, record):
        val = self.extract_data(record)
        self.register_col_width(val)
        self.sheet.awrite(val, self.format_data(val))

    def write_total(self):
        self.sheet.awrite('', self.format_total())


class SheetPortraitColumn(SheetColumn):
    """
        Column for ReportPortraitSheet. Not tied to a specific data field
        (as that will change from row to row), but tracks column width needed
    """
    def __init__(self, record=None, *args, **kwargs):
        super(SheetPortraitColumn, self).__init__(*args, **kwargs)
        self.xls_computed_width = 0
        self.record = record

    def write_data(self, data_or_field, fmt=None):
        val = self.fetch_val(self.record, data_or_field)
        if val is not None:
            self.register_col_width(val)
        self.sheet.awrite(val, fmt)


class BlankColumn(SheetColumn):
    def extract_data(self, record):
        return ''


class LabeledColumn(SheetColumn):
    header_start_row = 0

    def __init__(self, label=None, key=None, **kwargs):
        if label:
            label_rows = label.split('\n')
            for i_row, label_string in enumerate(label_rows):
                kwargs['header_{0}'.format(self.header_start_row + i_row)] = label_string
        super(LabeledColumn, self).__init__(key=key, **kwargs)


class PortraitRow(SheetUnit):
    """A row object for a portrait sheet with multiple columns
    The arguments here match the columns created in the ReportPortraitSheet instance's
    `init_columns`. Since this report is in
    Portrait mode this is necessary to display the report properly.
    In the end the report will look like this:
    +--------------+---------+
    | column_1     | val_1   | <- Not rendered, it's just here for a visual aid.
    +--------------+---------+
    | Foo          | Bar     | PortraitRow('Foo', Model.foo)
    +--------------+---------+
    | Baz          | 0.1145  | PortraitRow('Baz', Model.display_value)
    +--------------+---------+
    | ...          | ...     | Other fields...
    +--------------+---------+
    """

    def __init__(self, sheet=None, render_func=None, **kwargs):
        """Instantiation of the row
        NOTE: Render func is used for overriding rendering functionality
            * Instead of baking in a default render and then having to subclass this object you can
            pass in `render_func` which takes a single argument, the PortraitRow instance. Call
            `write_data()` on the columns as you see fit. If you return True or None `render` will
            move to the next row, return `False` to force tribune to stay on the same row.
            def render_func(row):
                row.sheet.columns['label'].write_data(row.label, FORMAT_DICTIONARY)
                row.sheet.columns['value'].write_data(row.value, FORMAT_DICTIONARY)
        """
        self.sheet = sheet
        self.render = render_func or self.render
        self.column_format = {
            'label': {},
            'value': {},
        }

    def render(self):
        raise NotImplementedError

    def update_format(self, label={}, value={}):
        self.column_format['label'].update(label)
        self.column_format['value'].update(value)


class TotaledMixin(object):
    def write_total(self):
        sum_begin = self.sheet.pre_data_rows+1
        sum_end = self.sheet.rownum
        if sum_begin > sum_end:
            # no data rows
            self.sheet.awrite('', self.format_total())
            return
        self.sheet.awrite(
            '=SUM({0}{1}:{0}{2})'.format(
                column_letter(self.sheet.colnum),
                sum_begin,
                sum_end
            ),
            self.format_total()
        )


class SheetSection(SheetUnit):
    """
        Groups SheetColumns together. May override heading render for
        individual columns (e.g. need a merged cell to head the section)

        If a sheet references a section, it will not reference the individual
        columns directly but depend on the section to police its own
    """
    __metaclass__ = _SheetDeclarativeMeta
    __cls_units__ = ()

    def new_instance(self, sheet):
        section = copy.deepcopy(self)
        section.sheet = sheet
        section.units = []
        for unit in self.__cls_units__:
            new_unit = unit.new_instance(section.sheet)
            section.units.append(new_unit)
        return section

    def write_header(self):
        # track starting column and then delegate to subunits
        self.col_idx = self.sheet.colnum
        for u in self.units:
            u.write_header()

    def write_data(self, record):
        # delegate to subunits
        for u in self.units:
            u.write_data(record)

    def write_total(self):
        # delegate to subunits
        for u in self.units:
            u.write_total()

    def adjust_col_width(self):
        # delegate to subunits
        for u in self.units:
            u.adjust_col_width()


class ReportSheet(WriterX, SheetSection):
    freeze = None
    pre_data_rows = 0
    sheet_name = None

    def __init__(self, parent_book, worksheet=None, **kwargs):
        self.parent_book = parent_book
        if not worksheet:
            worksheet = parent_book.add_worksheet(self.sheet_name)
        super(ReportSheet, self).__init__(ws=worksheet)
        self.set_base_style_dicts()

        # fetch records first, as units may need some data for initialization
        filter_args = dict([
            (k[7:], v) for k, v in six.iteritems(kwargs) if k[:7] == 'filter_'
        ])
        self.records = self.fetch_records(**filter_args)

        # list of (row,col) tuples to guard when writing. Idea here is to have
        #   as short a list as possible, for performance (i.e. don't check
        #   xlwt's complete sheet or row list)
        self.locked_cells = []

        # generate list of units (whether basic units or sections) from declarative data
        self.units = []
        for unit in self.__cls_units__:
            new_unit = unit.new_instance(self)
            self.units.append(new_unit)

        if kwargs.get('auto_render', True):
            self.render()

    def colwidth(self, chars):
        return chars * 256

    def fontsize(self, pt):
        return pt * 20

    def add_format_to_workbook(self, fmt):
        # attach a Format object to the workbook, if it has not been already
        if fmt.xf_format_indices is None:
            fmt.xf_format_indices = self.parent_book.xf_format_indices
            fmt.dxf_format_indices = self.parent_book.dxf_format_indices
            self.parent_book.formats.append(fmt)

    def set_base_style_dicts(self):
        self.style_title = {'bold': True, 'font_size': 14}
        self.style_header = {'bold': True}
        self.style_row_header = {'bold': True, 'bottom': 1}
        self.style_row_pct = {'num_format': '0.00%'}
        self.style_row_pct_round = {'num_format': '0%'}
        self.style_row_totals = {'bold': True, 'top': 1}
        self.style_row_totals_pct = {'bold': True, 'top': 1, 'num_format': '0.00%'}
        self.style_italic = {'italic': True}
        self.style_date_mdy = {'num_format': 'M/D/YYYY'}
        self.style_int = {'num_format': '#,##0'}
        self.style_numeric = {'num_format': '#,##0.00'}
        self.style_pct = {'num_format': '0.00%'}
        self.style_left = {'align': 'left'}
        self.style_center = {'align': 'center'}
        self.style_right = {'align': 'right'}
        self.style_currency = {'num_format': '$* #,##0.00;[Red]$* (#,##0.00);$* -  ;'}

    def combine_style_dicts(self, *args):
        # takes dicts such as those defined in set_base_styles, combines
        #   left-to-right, returns dict
        def add_dicts(a, b):
            return dict(list(a.items()) + list(b.items()) +
                        [(k, a[k] + b[k]) for k in set(b) & set(a)])

        style = args[0]
        if len(args) == 1:
            return style
        iterargs = iter(args)
        next(iterargs)
        for s in iterargs:
            style = add_dicts(style, s)
        return style

    def fetch_records(self, **kwargs):
        raise NotImplementedError

    def freeze_pane(self, col, row):
        self.ws.freeze_panes(row, col)

    def write_simple_merge(self, num_cols, data, style=None):
        """shorthand for WriterX.write_merge, for merge on single row
            data: can be a literal value, or a tuple of (record, key) which will do the fetch_val
        """
        data_to_write = data
        if isinstance(data, tuple):
            # extract here to raise an exception if the tuple doesn't have the right number of vals
            record, key = data
            data_to_write = _fetch_val(record, key)
        if num_cols == 0:
            raise Exception('Cannot write data length 0')
        if num_cols > 1:
            self.ws.merge_range(self.rownum, self.colnum, self.rownum, self.colnum + num_cols - 1,
                                data_to_write, self.conform_style(style))
        else:
            # number of cells is one, so no merge needed
            self.write(self.rownum, self.colnum, data_to_write, self.conform_style(style))
        self.colnum += num_cols

    def write_sheet_header(self):
        # use to write sheet header info which is not column-specific (e.g. title)
        pass

    def write_header(self, row, col, data, style=None):
        # calls write, and records row,col tuple to guard overwrites
        self.write(row, col, data, style)
        self.locked_cells.append((row, col))

    def lock_sheet_header(self):
        # should result in performance upgrade, constant time lookup on locked_cells
        self.locked_cells = set(self.locked_cells)

    def conform_style(self, style):
        if isinstance(style, dict):
            # styles can be either an xlsxwriter Format or a dict
            style = Format(style)
            self.add_format_to_workbook(style)
        return style

    def write(self, row, col, data, style=None):
        """wraps WriterX.write, checks row,col tuple against guarded cells
            data: can be a literal value, or a tuple of (record, key) which will do the fetch_val
        """
        data_to_write = data
        if isinstance(data, tuple):
            # extract here to raise an exception if the tuple doesn't have the right number of vals
            record, key = data
            data_to_write = _fetch_val(record, key)
        if (row, col) not in self.locked_cells:
            self.ws.write(row, col, data_to_write, self.conform_style(style))

    def awrite(self, data, style=None, nextrow=False):
        super(ReportSheet, self).awrite(data, style=self.conform_style(style), nextrow=nextrow)

    def render_header(self):
        for i in range(self.pre_data_rows):
            for u in self.units:
                u.write_header()
            self.nextrow()

    def render_data(self):
        for r in self.records:
            for u in self.units:
                u.write_data(r)
            self.nextrow()

    def render_total(self):
        for u in self.units:
            u.write_total()
            u.adjust_col_width()

    def render(self):
        self.write_sheet_header()
        self.lock_sheet_header()
        self.render_header()
        self.render_data()
        self.render_total()
        if self.freeze:
            self.freeze_pane(*self.freeze)


class ReportPortraitSheet(ReportSheet):
    # generic reporting sheet in which data is laid out in rows, though perhaps
    #   with multiple columns providing depth

    def __init__(self, *args, **kwargs):
        kwargs['auto_render'] = False
        super(ReportPortraitSheet, self).__init__(*args, **kwargs)

        self.columns = self.init_columns()

        # assign the columns their indices. columns should be OrderedDict, so
        #   columns will come in the order designed
        for idx, c in enumerate(self.columns.values()):
            c.col_idx = idx

        self.render()

    def init_columns(self):
        # return an ordered dict of columns on the report
        raise NotImplementedError

    def render(self):
        self.write_sheet_header()
        self.lock_sheet_header()

        # reset location for data loop
        self.rownum = self.pre_data_rows
        self.colnum = 0

        # each unit is responsible for rendering all columns for its row
        #   if the unit's render returns True, this indicates it has rendered
        #   and the sheet can move to the next row
        for u in self.units:
            if u.render():
                self.nextrow()

        # columns have tracked the necessary widths, set here
        for c in self.columns.values():
            c.adjust_col_width()



from blazeutils.strings import randchars
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base

from .utils import DefaultMixin

meta = sa.MetaData()
session = None
Base = declarative_base(metadata=meta)


class Person(Base, DefaultMixin):
    __tablename__ = 'persons'

    id = sa.Column(sa.Integer, primary_key=True)
    firstname = sa.Column(sa.String(50))
    lastname = sa.Column('last_name', sa.String(50))
    intcol = sa.Column(sa.Integer)
    floatcol = sa.Column(sa.Float)

    def __repr__(self):
        return '<Person: "%s, created: %s">' % (self.id, self.createdts)

    @classmethod
    def testing_create(cls, **kwargs):
        kwargs['firstname'] = kwargs.get('firstname') or randchars()
        return cls.add(**kwargs)



"""A vocabulary for declaratively describing how to import sheets.

Example:

    from tribune.sheet_import import Field, NestedItersSheetData, SheetImporter
    from tribune.sheet_import.parsers import (
        chain, parse_int, parse_text, parse_yes_no, validate_min
    )

    class MySheet(SheetImporter):
        fields = (
            Field('Name', 'name',    parse_text),
            Field('Age',  'age',     chain(parse_int, validate_min(1))),
            Field('Male', 'is_male', parse_yes_no),
        )

    sheet = NestedItersSheetData([
        ('Name',           'Age', 'Male'),
        ('Yuri Gagarin',   '34',  'Yes'),
        ('Alexei Leonov',  '81',  'Yes'),
        ('Amelia Earhart', '39',  'No'),
    ])
    records = MySheet().get_sheet_records(sheet)
"""
import abc
from collections import namedtuple
from io import BytesIO

from blazeutils.spreadsheets import xlsx_to_reader
from six import with_metaclass, text_type
from six.moves import filter
import xlrd
from xlsxwriter import Workbook

import tribune

from ..utils import (
    column_letter,
    flatten,
    raises,
    split_every,
    unzip,
)


class SpreadsheetImportError(Exception):
    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        return '<SpreadsheetImportError: {}>'.format(self.errors)


def workbook_safe_open(file_handle):
    try:
        return xlrd.open_workbook(file_contents=file_handle.read())
    except xlrd.biffh.XLRDError as e:
        msg = str(e)
        # we don't want to screen all xlrd load errors, but a few common ones need to be trapped
        #   for user-friendliness
        if 'unsupported format' in msg.lower() or 'not supported' in msg.lower() \
                or 'not a known type of workbook' in str(e).lower():
            raise SpreadsheetImportError([u'File format is not recognized. Please upload an Excel'
                                          ' file (.xlsx or .xls).'])
        elif 'file size is 0 bytes' in msg.lower():
            raise SpreadsheetImportError([u'File is empty.'])
        else:
            raise SpreadsheetImportError([msg])


def normalize_text(x):
    """Normalizes a string by stripping whitespace and flattening case."""
    return text_type(x).strip().lower()


class Cell(object):
    """Represents a particular cell on a spreadsheet."""
    def __init__(self, row, column):
        """
        :param row: is a 0-based row index.
        :param column: is a 0-based column index.
        """
        self.row = row
        self.column = column

    def __str__(self):
        return column_letter(self.column) + str(self.row + 1)


class SheetData(with_metaclass(abc.ABCMeta)):
    """An ABC that describes the minimum necessary API for importing a sheet."""
    @abc.abstractmethod
    def __len__(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def cell_value(self, cell):
        raise NotImplementedError()


class XlrdSheetData(SheetData):
    def __init__(self, sheet):
        self.sheet = sheet

    def __len__(self):
        return self.sheet.nrows

    def cell_value(self, cell):
        try:
            return self.sheet.cell_value(cell.row, cell.column)
        except IndexError:
            return ''

    @classmethod
    def from_nested_iters(cls, data):
        """Builds an XlrdSheetData from nested iterables.

        param data: is nested iterables representing data for each cell.
                    For example, `[('A1', 'B1'), ('A2', 'B2')]` represents two rows and two columns
                    of data.
        """
        workbook = Workbook(BytesIO())
        sheet = workbook.add_worksheet('test')

        for row, row_data in enumerate(data):
            for column, cell_data in enumerate(row_data):
                sheet.write(row, column, cell_data)

        return XlrdSheetData(xlsx_to_reader(workbook).sheet_by_index(0))


class NestedItersSheetData(SheetData):
    """Wraps nested iterables to provide the SheetData API."""

    def __init__(self, data):
        """
        param data: is nested iterables representing data for each cell.
                    For example, `[('A1', 'B1'), ('A2', 'B2')]` represents two rows and two columns
                    of data.
        """
        self.data = data

    def __len__(self):
        return len(self.data)

    def cell_value(self, cell):
        try:
            return self.data[cell.row][cell.column]
        except IndexError:
            return ''


ErrorOrValue = namedtuple('ErrorOrValue', ['error', 'value'])


def aggregate_error_or_value(error_or_values):
    """Aggregates a list of ErrorOrValue tuples into a single ErrorOrValue (very much like `unzip`).
    If there any errors the resulting ErrorOrValue will only have list of errors.
    If there are no errors the resulting ErrorOrValue will have a list of values.
    """
    raw_errors, values = list(unzip(error_or_values)) or ([], [])
    errors = [error for error in flatten(raw_errors) if error is not None]
    return ErrorOrValue(errors, None) if errors else ErrorOrValue([], values)


Field = namedtuple('Field', ['label', 'field', 'parser'])


class SheetImporter(object):
    """A base class for defining how to parse a single spreadsheet."""
    data_start_row = 1  # 0-based index where first row of data begins
    fields = ()         # Ordered iterable of Field tuples

    def __init__(self, fields=None, data_start_row=None):
        """
        :param fields: is an ordered iterable of Field tuples. If not provided, the class-level
                       value will be used.
        :param data_start_row: is the 0-based index where the first row of data begins. If not
                               provided, the class-level value will be used.
        """
        self.fields = fields or self.fields
        self.data_start_row = data_start_row or self.data_start_row

    def _get_header_errors(self, sheet):
        """Builds a list of errors regarding the sheet header. The list will be empty if there are
        no errors."""
        header = self.data_start_row - 1
        return [
            'Expected "{header}" in header cell {cell}.'.format(
                header=field.label,
                cell=Cell(header, col)
            ) for col, field in enumerate(self.fields)
            if normalize_text(sheet.cell_value(Cell(header, col))) != normalize_text(field.label)
        ]

    @staticmethod
    def _parse_cell(field, cell, value):
        """Parses a cell using the given field definition. The result is an ErrorOrValue tuple.

        :param field: is a Field tuple which defines the field.
        :param cell: is a Cell which contains both row and column indexes.
        :param value: is the raw value at that cell in the sheet.
        """
        try:
            return ErrorOrValue([], field.parser(value))
        except SpreadsheetImportError as e:
            contextualized_errors = [
                'Unexpected value in cell {cell}: {msg}'.format(cell=cell, msg=msg)
                for msg in e.errors
            ]
            return ErrorOrValue(contextualized_errors, None)

    def _parse_row(self, sheet, row):
        """Parses all the columns in a sheet at a given row index and returns aggregated errors or
        values (for every column) as an ErrorOrValue tuple."""
        col_data = [self._parse_cell(field, Cell(row, col), sheet.cell_value(Cell(row, col)))
                    for col, field in enumerate(self.fields)]
        errors, values = aggregate_error_or_value(col_data)
        return ErrorOrValue(errors, None) if errors \
            else ErrorOrValue([], self.row_to_record(values))

    def _parse_sheet_data(self, sheet):
        """Parses only the data rows from a sheet and returns aggregated errors or valuse as an
        ErrorOrValue tuple."""
        return [self._parse_row(sheet, row) for row in range(self.data_start_row, len(sheet))]

    def _parse_sheet(self, sheet):
        """Parses a sheet (including the headers) and returns aggregated errors or values as an
        ErrorOrValue tuple."""
        header_errors = self._get_header_errors(sheet)
        if header_errors:
            return [ErrorOrValue(header_errors, None)]
        return self._parse_sheet_data(sheet)

    def row_to_record(self, values):
        """Converts a list of values from a row into a dictionary where each value is keyed
        on its column name from `self.fields`."""
        return self.modify_record({field.field: v for field, v in zip(self.fields, values)})

    def modify_record(self, record):
        """Override to add/remove/change fields in the record (dictionary of column to value).

        :returns: the record with any modifications applied or None to filter out the record
                  entirely.
        """
        return record

    def get_sheet_records(self, sheet):
        """Parses a sheet and returns an iterator of rows where each row is a dictionary of data
        values keyed by their column name. If errors occurred during the parse, this aggregates them
        and raises them in `SpreadsheetImportError`."""
        errors_or_values = self._parse_sheet(sheet)
        errors, records = aggregate_error_or_value(errors_or_values)
        # Filter out empty records.
        return filter(None, records) if not errors else raises(SpreadsheetImportError(errors))

    def get_sheet_records_batched(self, sheet, batch_size):
        """Like `get_sheet_records` but returns a generator that splits the result into batches."""
        return split_every(batch_size, self.get_sheet_records(sheet))

    def as_report_sheet(self):
        """Returns a ReportSheet class which mirrors this SheetImporter."""
        class NewReportSheet(tribune.ReportSheet):
            pre_data_rows = self.data_start_row

            for ___field in self.fields:  # This iteration variable gets left behind on the class!
                tribune.LabeledColumn(___field.label)

            def fetch_records(self):
                return []

        return NewReportSheet



import copy
import string
from itertools import (
    chain as _itertools_chain,
    count,
    islice,
    takewhile,
)
from functools import reduce

from six.moves import map


def raises(e):
    """A functional form of `raise`."""
    raise e


def compose(*functions):
    """Function composition on a series of functions.

    Remember that function composition runs right-to-left: `f . g . h = f(g(h(x)))`. As a unix
    pipeline, it would be written: `h | g | f`.

    From https://mathieularose.com/function-composition-in-python/.
    """
    return reduce(lambda f, g: lambda x: f(g(x)), functions, lambda x: x)


def chain(*functions):
    """Returns a new function based on a series of given functions by chaining them together via
    function composition. This is exactly equal to function composition, but represents the chaining
    from left to right instead of from right to left.

    Use this to combine parsers and validators.
    """
    return compose(*reversed(functions))


def unzip(iterable):
    """Unzips/transposes an iterable of tuples into a tuple of lists.

    WARNING: When given an empty iterable, this returns an empty list instead of a tuple. If you
    need a consistent interface then do something like this:

        left, right = list(unzip(two_columned_list)) or ([], [])

    :returns: an iterable of the unzipped input.
    """
    return map(list, zip(*iterable))


def flatten(iterable):
    """Takes an iterable of iterables and flattens it by one layer (e.g. [[1],[2]] becomes [1,2]).
    """
    return list(_itertools_chain(*iterable))


def deepcopy_tuple_except(tup, exceptions=type(None)):
    """
    Deeply-copy a nested tuple, but don't copy any items within the
    tuple which have a type listed in `exceptions`

    :param tup: the tuple to copy
    :param exceptions: a set of types not to copy
    :return: a clone of `tup`
    """
    result = []

    for item in tup:
        if isinstance(item, exceptions):
            result.append(item)
        else:
            if isinstance(item, tuple):
                result.append(deepcopy_tuple_except(item, exceptions))
            else:
                result.append(copy.deepcopy(item))

    return tuple(result)


def split_every(n, iterable):
    """Returns an iterable that spits an iterable into n-sized chunks. The last chunk may have less
    than n elements.

    See http://stackoverflow.com/a/22919323/503377."""
    items = iter(iterable)
    return takewhile(bool, (list(islice(items, n)) for _ in count(0)))


def column_letter(n):
    """Returns the spreadsheet column name for a given 0-based column index.

    Adapted from http://stackoverflow.com/a/4532562/503377. Spreadsheet column names are a strange
    base-26 number (for 26 letters in the alphabet) where each place value is 1-based instead of
    0-based.

    :param n: is a 0-based column index.
    """
    alphabet = string.ascii_uppercase

    n = n + 1  # Turn our 0-based number into a 1-based number.
    name = ''  # initial column name
    while n > 0:
        n = n - 1  # Turn this place value into a 0-based number.
        name = alphabet[n % len(alphabet)] + name
        n //= len(alphabet)
    return name



import datetime as dt

import six
import sqlalchemy as sa
import sqlalchemy.orm as saorm
import sqlalchemy.sql as sasql
import wrapt

from tribune.sheet_import import SpreadsheetImportError


# utility functions to find a row or column for testing. Reduces need for
#   hard-coded row/col indices in tests.
#   call generally with the worksheet, desired cell value, the column/row to search
#       within, and starting/ending indices
#   returns -1 if the value is not found

def find_sheet_row(sheet, needle, column, start=0, end=100):
    try:
        for x in range(start, end+1):
            if sheet.cell_value(x, column) == needle:
                return x
    except IndexError:
        pass
    return -1


def find_sheet_col(sheet, needle, row, start=0, end=100):
    try:
        for x in range(start, end+1):
            if sheet.cell_value(row, x) == needle:
                return x
    except IndexError:
        pass
    return -1


def assert_import_errors(errors, func):
    try:
        func()
        assert False, 'Did not raise SpreadsheetImportError'
    except SpreadsheetImportError as e:
        assert set(e.errors) == set(errors)


class DefaultColsMixin(object):
    id = sa.Column(sa.Integer, primary_key=True)
    createdts = sa.Column(sa.DateTime, nullable=False, default=dt.datetime.now,
                          server_default=sasql.text('CURRENT_TIMESTAMP'))
    updatedts = sa.Column(sa.DateTime, nullable=False, default=dt.datetime.now,
                          server_default=sasql.text('CURRENT_TIMESTAMP'), onupdate=dt.datetime.now)


@wrapt.decorator
def transaction(f, decorated_obj, args, kwargs):
    """
        decorates a function so that a DB transaction is always committed after
        the wrapped function returns and also rolls back the transaction if
        an unhandled exception occurs.

        'ncm' = non class method (version)
    """
    from tribune.tests.entities import session as dbsess

    try:
        retval = f(*args, **kwargs)
        dbsess.commit()
        return retval
    except Exception:
        dbsess.rollback()
        raise


def transaction_classmethod(f):
    """
        like transaction() but makes the function a class method
    """
    return transaction(classmethod(f))


class MethodsMixin(object):
    @classmethod
    def _sa_sess(cls):
        from .entities import session
        return session

    @transaction_classmethod
    def add(cls, **kwargs):
        o = cls()
        o.from_dict(kwargs)
        cls._sa_sess().add(o)
        return o

    def from_dict(self, data):
        """
        Update a mapped class with data from a JSON-style nested dict/list
        structure.
        """
        # surrogate can be guessed from autoincrement/sequence but I guess
        # that's not 100% reliable, so we'll need an override

        mapper = saorm.object_mapper(self)

        for key, value in six.iteritems(data):
            if isinstance(value, dict):
                dbvalue = getattr(self, key)
                rel_class = mapper.get_property(key).mapper.class_
                pk_props = rel_class._descriptor.primary_key_properties

                # If the data doesn't contain any pk, and the relationship
                # already has a value, update that record.
                if not [1 for p in pk_props if p.key in data] and \
                   dbvalue is not None:
                    dbvalue.from_dict(value)
                else:
                    record = rel_class.update_or_create(value)
                    setattr(self, key, record)
            elif isinstance(value, list) and \
                    value and isinstance(value[0], dict):

                rel_class = mapper.get_property(key).mapper.class_
                new_attr_value = []
                for row in value:
                    if not isinstance(row, dict):
                        raise Exception(
                            'Cannot send mixed (dict/non dict) data '
                            'to list relationships in from_dict data.'
                        )
                    record = rel_class.update_or_create(row)
                    new_attr_value.append(record)
                setattr(self, key, new_attr_value)
            else:
                setattr(self, key, value)


class DefaultMixin(DefaultColsMixin, MethodsMixin):
    pass



from decimal import Decimal as D


def pytest_configure(config):
    from sqlalchemy import engine_from_config
    from sqlalchemy.orm import sessionmaker, scoped_session
    engine = engine_from_config({'url': 'sqlite:///'}, prefix='')
    session = scoped_session(sessionmaker(bind=engine))

    from tribune.tests import entities
    entities.meta.bind = engine
    entities.meta.create_all()
    entities.session = session
    for x in range(1, 50):
        p = entities.Person()
        p.firstname = 'fn%03d' % x
        p.lastname = 'ln%03d' % x
        p.sortorder = x
        p.numericcol = D('29.26') * x / D('.9')
        session.add(p)

    session.commit()



import operator

from blazeutils.datastructures import OrderedDict

from tribune import (
    BlankColumn,
    LabeledColumn,
    PortraitRow,
    ReportPortraitSheet,
    ReportSheet,
    SheetPortraitColumn,
    SheetSection,
    TotaledMixin,
)

"""
    Objects for testing normal landscape-layout reports
"""


class CarLabeledColumn(LabeledColumn):
    header_start_row = 2


class CarYearColumn(CarLabeledColumn):
    xls_width = 10


class CarTotaledColumn(TotaledMixin, CarLabeledColumn):
    pass


class CarModelSection(SheetSection):
    def __init__(self, *args, **kwargs):
        self.foo = kwargs.get('foo')

    CarYearColumn('Year', 'year', xls_width=15)
    CarLabeledColumn('Make', 'make')
    CarLabeledColumn('Model', 'model')


class CarStyleSection(SheetSection):
    CarLabeledColumn('Style', 'style')
    CarLabeledColumn('Color', 'color')


class CarSheet(ReportSheet):
    pre_data_rows = 4
    sheet_name = 'Car Info'

    BlankColumn()
    CarModelSection(foo='bar')
    CarStyleSection()
    CarTotaledColumn('Blue\nBook', 'book_value')

    def fetch_records(self, **kwargs):
        # gets mocked in tests to return correct sequences
        pass

    def write_sheet_header(self):
        self.write_header(0, 0, 'Cars', self.style_header)


"""
    Objects for testing portrait-layout reports
"""


class DealershipReportRow(PortraitRow):
    def __init__(self, label=None, comp=None, inc=None, exp=None, net=None,
                 comp_curr=False, **kwargs):
        super(DealershipReportRow, self).__init__()
        self.label = label
        self.comp = comp
        self.comp_curr = comp_curr
        self.inc = inc
        self.exp = exp
        self.net = net

        if self.net is None and (self.inc or self.exp):
            self.net = (self.inc, self.exp, operator.sub)

    def format_label(self):
        return None

    def format_component(self):
        if self.comp_curr:
            return self.sheet.style_currency
        return None

    def format_income(self):
        return self.sheet.style_currency

    def format_expense(self):
        return self.sheet.style_currency

    def format_net(self):
        return self.sheet.style_currency

    def render(self):
        self.sheet.columns['label'].write_data(self.label, self.format_label())
        self.sheet.columns['component'].write_data(self.comp, self.format_component())
        self.sheet.columns['income'].write_data(self.inc, self.format_income())
        self.sheet.columns['expense'].write_data(self.exp, self.format_expense())
        self.sheet.columns['net'].write_data(self.net, self.format_net())
        return True


class DealershipTotalRow(DealershipReportRow):
    def format_label(self):
        return self.sheet.combine_style_dicts(
            self.sheet.style_header,
            self.sheet.style_right,
            {'top': 6}
        )

    def format_component(self):
        return self.sheet.combine_style_dicts(
            self.sheet.style_currency,
            {'top': 6}
        )

    def format_income(self):
        return self.format_component()

    def format_expense(self):
        return self.format_component()

    def format_net(self):
        return self.format_component()


class CarDealerSheet(ReportPortraitSheet):
    pre_data_rows = 4
    sheet_name = 'Dealership Summary'

    DealershipReportRow('Sales', 'sale_count', 'sale_income', 'sale_expense')
    DealershipReportRow('Rentals', 'rental_count', 'rental_income', 'rental_expense')
    DealershipReportRow('Leases', 'lease_count', 'lease_income', 'lease_expense')
    DealershipTotalRow('Totals', None, None, None, 'total_net')

    def fetch_records(self, **kwargs):
        # gets mocked in tests to return correct sequences
        pass

    def init_columns(self):
        d = OrderedDict()
        d['label'] = SheetPortraitColumn(sheet=self, record=self.records[0])
        d['component'] = SheetPortraitColumn(sheet=self, record=self.records[0])
        d['income'] = SheetPortraitColumn(sheet=self, record=self.records[0])
        d['expense'] = SheetPortraitColumn(sheet=self, record=self.records[0])
        d['net'] = SheetPortraitColumn(sheet=self, record=self.records[0])
        return d

    def write_sheet_header(self):
        self.write_header(0, 0, 'Dealership', self.style_header)
        self.write_header(3, 1, 'Count', self.style_header)
        self.write_header(3, 2, 'Income', self.style_header)
        self.write_header(3, 3, 'Expense', self.style_header)
        self.write_header(3, 4, 'Net', self.style_header)



from six import text_type
from six.moves import (
    filter,
    map,
)

from ..utils import (
    chain,
    raises,
)
from . import (
    SpreadsheetImportError,
    normalize_text,
)


def map_(func):
    """A strict (non-iterator), curried version of map for convenience in parser chains."""
    return lambda values: list(map(func, values))


def filter_(func):
    """A strict (non-iterator), curried version of filter for convenience in parser chains."""
    return lambda values: list(filter(func, values))


def parse_text(value):
    """Parses text as a unicode string and strips leading/trailing whitespace."""
    return text_type(value).strip()


def parse_yes_no(value):
    """Parses boolean values represented as 'yes' or 'no' (case insensitive)."""
    result = {
        normalize_text('yes'): True,
        normalize_text('no'):  False
    }.get(normalize_text(value))
    if result is None:
        raise SpreadsheetImportError([u'must be "yes" or "no"'])
    return result


def parse_int(value):
    """Parses integers. It accepts floating-point representations as long as they can be
    safely transformed into integers without loss of precision."""
    try:
        as_float = float(value)
        as_int = int(as_float)
        # compare float and integer versions because, e.g. 3.0 == 3 but 3.1 != 3.
        return as_int if as_int == as_float else raises(ValueError())
    except (ValueError, TypeError):
        raise SpreadsheetImportError([u'must be an integer'])


def parse_int_as_text(value):
    """Parses integers but converts them to unicode strings in the result."""
    return text_type(parse_int(value))


def parse_number(value):
    """Parses any floating point number."""
    try:
        return float(value)
    except (ValueError, TypeError):
        raise SpreadsheetImportError([u'must be a number'])


def parse_int_or_text(value):
    """Tries to parse a value as an integer and turn it into text but if that parse fails, it
    reverts to simple text parsing."""
    try:
        return parse_int_as_text(value)
    except SpreadsheetImportError:
        return parse_text(value)


def default_to(default):
    """Returns a parser that gives the provided default if the value is empty."""
    return lambda value: default if value is None or text_type(value).strip() == '' else value


def nullable(parser):
    """Returns a parser that wraps another parser and only applies it if the value is not empty.
    If the value is empty, this parser always converts it to `None`."""
    return lambda value: None if default_to(None)(value) is None else parser(value)


def validate_satisfies(pred, error_message):
    """Validates that a value satisfies the given predicate or issues the given error if it
    doesn't."""
    return lambda value: (value if pred(value)
                          else raises(SpreadsheetImportError([error_message])))


def validate_max_length(maximum):
    """Validates that the value has at most ``maximum`` digits."""
    return validate_satisfies(lambda v: len(v) <= maximum,
                              u'must be no more than {} characters long'.format(maximum))


validate_not_empty = chain(
    default_to(None),
    validate_satisfies(lambda v: v is not None, u'must not be empty')
)


def validate_min(min_):
    """Validates that a number is equal to or grater than the given minimum."""
    return validate_satisfies(lambda v: v >= min_, u'number must be no less than {}'.format(min_))


def validate_max(max_):
    """Validates that a number is equal to or less than the given maximum."""
    return validate_satisfies(lambda v: v <= max_,
                              u'number must be no greater than {}'.format(max_))


def validate_range(min_, max_):
    """Validates that a number is between a minimum and a maximum."""
    return chain(validate_max(max_), validate_min(min_))


def validate_unique(others, thing='thing'):
    """Validates that a value is not a member of the given set.

    :param others: is the other things which already exist.
    :param thing: is the name of the "thing" which may or may not be taken.
    """
    others_set = frozenset(others)

    def parser(value):
        if value in others_set:
            raise SpreadsheetImportError([
                u'{thing} "{value}" already exists'.format(thing=thing, value=value)])
        return value
    return parser


def validate_one_of(choices, thing='thing', show_choices_in_error=False):
    """Validates that a value is a member of the given set of choices.

    :param choices: is the set of possible choices.
    :param thing: is the name of the "things" that make up the possible choices.
    :param show_choices_in_error: determines if all the choices should be rendered in the error
                                  message.
    """
    error_message = u'must be a valid {thing}{choices}'.format(
        thing=thing,
        choices=': ' + ', '.join(map(str, choices)) if show_choices_in_error else ''
    )
    choices_set = frozenset(choices)
    return validate_satisfies(lambda v: v in choices_set, error_message)


def parse_list(parser=lambda x: x, delim=' '):
    """Returns a parser that splits its input text by the given delimiter and runs the given parser
    on each element of the result.

    :param parser: is a parser to map onto each element in the parsed list. By default it is the
                   identity function.
    :param delim: is the delimiter to use for splitting the input into a list.
    """
    return chain(parse_text, lambda value: value.split(delim), map_(parser))


# Keeps only non-empty values in a list
filter_non_empty = chain(map_(default_to(None)), filter_(lambda x: x is not None))


def parse_lookup(mapping, thing='thing'):
    """Parses a value by looking it up in a mapping (usually a dictionary). If the value is missing,
    this raises a SpreadsheetImportError. If you want to supply a default instead, then provide a
    `collections.defaultdict` object as the mapping.

    :param mapping: is a mapping from raw value to desired value. It only needs to behave like a
                    `dict`.
    :param thing: is the name of the resulting type for error messages.
    """
    def parser(value):
        try:
            return mapping[value]
        except KeyError:
            raise SpreadsheetImportError([u'"{}" is not a valid {}'.format(value, thing)])

    return parser



import pytest

from tribune.utils import (
    column_letter,
    deepcopy_tuple_except,
    split_every,
)


def test_column_letter():
    assert column_letter(0) == 'A'
    assert column_letter(2) == 'C'
    assert column_letter(25) == 'Z'
    assert column_letter(26) == 'AA'
    assert column_letter(27) == 'AB'
    assert column_letter(26*2) == 'BA'
    assert column_letter(26*3) == 'CA'
    assert column_letter(26**2+25) == 'ZZ'
    assert column_letter(26**2+26) == 'AAA'


class TestDeepCopyTupleExcept(object):
    class DoDeepCopy:
        def __init__(self, value):
            self.value = value

        def __eq__(self, other):
            return self.value == other.value

    class DoNotDeepCopy(DoDeepCopy):
        def __deepcopy__(self, memodict={}):
            raise Exception('Deep Copy!')

    def test_simple(self):
        original = (1, 2, 3)
        clone = deepcopy_tuple_except(original)
        assert original == clone

    def test_deep(self):
        original = (1, (2, self.DoNotDeepCopy([3, 4, 5]), 6), 7)
        clone = deepcopy_tuple_except(original, self.DoNotDeepCopy)
        assert original == clone

    def test_clone_is_clone(self):
        original = (1, [2, 3, 4], 5)
        clone = deepcopy_tuple_except(original)
        original[1].append(99)
        assert original != clone
        assert clone[0] == 1
        assert clone[1] == [2, 3, 4]
        assert clone[2] == 5

    def test_objects_are_deeply_copied(self):
        original = (1, self.DoDeepCopy([2, 3, 4]), 5)
        clone = deepcopy_tuple_except(original)
        assert original == clone

        assert clone[1].value == [2, 3, 4]
        original[1].value.append(99)
        assert clone[1].value == [2, 3, 4]

    def test_exceptions_are_not_deeply_copied(self):
        original = (1, self.DoNotDeepCopy([2, 3, 4]), 5)
        clone = deepcopy_tuple_except(original, self.DoNotDeepCopy)
        assert original == clone

        assert clone[1].value == [2, 3, 4]
        original[1].value.append(99)
        assert clone[1].value == [2, 3, 4, 99]

        assert original == clone

    def test_exceptions_are_not_copied_at_all(self):
        original = (1, self.DoNotDeepCopy([2, 3, 4]), 5)
        clone = deepcopy_tuple_except(original, self.DoNotDeepCopy)

        assert original[1] is clone[1]

    def test_exceptions_are_used(self):
        original = (1, self.DoNotDeepCopy([2, 3, 4]), 5)

        with pytest.raises(Exception):
            deepcopy_tuple_except(original)


def test_split_every():
    assert list(split_every(1, [])) == []
    assert list(split_every(1, [1])) == [[1]]
    assert list(split_every(1, [1, 2])) == [[1], [2]]
    assert list(split_every(2, [1, 2])) == [[1, 2]]
    assert list(split_every(3, [1, 2])) == [[1, 2]]
    assert list(split_every(2, [1, 2, 3])) == [[1, 2], [3]]



import os.path as osp
try:
    from setuptools import setup
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup

# pip install -e .[develop]
develop_requires = [
    'BlazeUtils',
    'SQLAlchemy',
    'XlsxWriter',
    'mock',
    'pytest',
    'pytest-cov',
    'six',
    'wrapt',
    'xlrd',
]

cdir = osp.abspath(osp.dirname(__file__))
README = open(osp.join(cdir, 'readme.rst')).read()
CHANGELOG = open(osp.join(cdir, 'changelog.rst')).read()

version_fpath = osp.join(cdir, 'tribune', 'version.py')
version_globals = {}
with open(version_fpath) as fo:
    exec(fo.read(), version_globals)

setup(
    name="tribune",
    version=version_globals['VERSION'],
    description="A library for coding Excel reports in a declarative fashion",
    long_description='\n\n'.join((README, CHANGELOG)),
    author="Matt Lewellyn",
    author_email="matt.lewellyn@level12.io",
    url='https://github.com/level12/tribune',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    license='BSD',
    packages=[
        'tribune',
        'tribune.sheet_import',
    ],
    extras_require={'develop': develop_requires},
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        'BlazeUtils',
        'six',
        'xlsxwriter',
    ],
)



import operator
from io import BytesIO

from blazeutils.datastructures import BlankObject
from blazeutils.spreadsheets import xlsx_to_reader
import mock
import pytest
from xlsxwriter import Workbook
import wrapt

from tribune import (
    LabeledColumn,
    PortraitRow,
    ProgrammingError,
    ReportSheet,
    SheetColumn,
    SheetSection,
    SheetUnit,
)
from .entities import Person
from .reports import CarSheet, CarDealerSheet
from .utils import find_sheet_col, find_sheet_row


def car_data(self):
    return [
        {'year': 1998, 'make': 'Ford', 'model': 'Taurus', 'style': 'SE Wagon',
         'color': 'silver', 'book_value': 1500},
        {'year': 2004, 'make': 'Oldsmobile', 'model': 'Alero', 'style': '4D Sedan',
         'color': 'silver', 'book_value': 4500},
        {'year': 2003, 'make': 'Ford', 'model': 'F-150', 'style': 'XL Supercab',
         'color': 'blue', 'book_value': 8500},
    ]


def dealer_data(self):
    return [
        {'sale_count': 5, 'sale_income': 75000, 'sale_expense': 25000,
         'rental_count': 46, 'rental_income': 8600, 'rental_expense': 2900,
         'lease_count': 27, 'lease_income': 10548, 'lease_expense': 3680,
         'total_net': 100000},
    ]


class TestSheetDecl(object):
    @mock.patch('tribune.tests.reports.CarSheet.fetch_records', car_data)
    def test_sheet_units(self):
        sheet = CarSheet(Workbook(BytesIO()))
        assert len(sheet.units) == 4

    @mock.patch('tribune.tests.reports.CarSheet.fetch_records', car_data)
    def test_section_units(self):
        sheet = CarSheet(Workbook(BytesIO()))
        section = sheet.units[1]
        assert len(section.units) == 3

    @mock.patch('tribune.tests.reports.CarSheet.fetch_records', car_data)
    def test_attribute_values_passed_to_section_instance(self):
        sheet = CarSheet(Workbook(BytesIO()))
        car_model_section = sheet.units[1]
        assert car_model_section.foo == 'bar'

    @mock.patch('tribune.tests.reports.CarSheet.fetch_records', car_data)
    def test_attribute_values_passed_to_column_instance(self):
        sheet = CarSheet(Workbook(BytesIO()))
        car_model_section = sheet.units[1]
        car_year_column = car_model_section.units[0]
        assert car_year_column.xls_width == 15

    @mock.patch('tribune.tests.reports.CarSheet.fetch_records')
    def test_filter_args(self, m_fetch):
        CarSheet(Workbook(BytesIO()), filter_arg_a='foo', filter_arg_b='bar')
        m_fetch.assert_called_once_with(arg_a='foo', arg_b='bar')

    def test_deep_copy_sqlalchemy(self):
        """SQLALchemy objects like InstrumentedAttribute do not like deep copies. In the cases used
        by tribune, anything inheriting QueryableAttribute is treated differently, simply assigning
        the value over on the copy (instead of copying and then copying all underlying objects)."""
        base_row = PortraitRow('foo', Person.firstname)
        base_row.new_instance(None)

    def test_deep_copy_sqlalchemy_tuple(self):
        """
        tribune supports tuple expressions as values (formulated as as a 3-tuple,
        `(value1, value2, operator)`). As the tuple-values could be QueryableAttributes,
        we don't necessarily want to deep-copy

        :return:
        """
        base_row = PortraitRow('foo', (Person.intcol, -1, operator.mul))
        base_row.new_instance(None)

        base_row = PortraitRow('foo', (-1, Person.intcol, operator.mul))
        base_row.new_instance(None)

        base_row = PortraitRow('foo', ((Person.intcol, -1, operator.mul), -1, operator.mul))
        base_row.new_instance(None)

    def test_deep_copy_unbound_method(self):
        """This test is kind of hard to set up, so it needs some explanation. A great deal of
        function objects are fine with the deep copy, but this particular case is not. What we
        have is an unbound method (status_render) wrapped in a decorator."""
        def status_render(row):
            pass

        @wrapt.decorator
        def some_decorator(wrapped, instance, args, kwargs):
            row, = args
            return wrapped(row)

        class DummyUnit(SheetUnit):
            def __init__(self):
                self.foo = some_decorator(status_render)

        base_unit = DummyUnit()
        base_unit.new_instance(None)

    def test_deep_copy_static_method(self):
        """This test is kind of hard to set up, so it needs some explanation. A great deal of
        function objects are fine with the deep copy, but this particular case is not. What we
        have is an static method (status_render) wrapped in a decorator."""
        @wrapt.decorator
        def some_decorator(wrapped, instance, args, kwargs):
            row, = args
            return wrapped(row)

        class DummyUnit(SheetUnit):
            @staticmethod
            def status_render(row):
                pass

            def __init__(self):
                self.foo = some_decorator(self.status_render)

        base_unit = DummyUnit()
        base_unit.new_instance(None)


class TestSheetColumn(object):
    def test_extract_data_no_key(self):
        col = SheetColumn()
        assert col.extract_data({}) == ''

    def test_extract_data_attr(self):
        col = SheetColumn('bar')
        assert col.extract_data(
            BlankObject(
                foo=1,
                bar='something',
                baz='else',
            )
        ) == 'something'

    def test_extract_data_dict(self):
        col = SheetColumn('bar')
        assert col.extract_data({'foo': 1, 'bar': 'something', 'baz': 'else'}) == 'something'

    def test_extract_data_string(self):
        col = SheetColumn('bar')
        assert col.extract_data({'foo': 1, 'baz': 'else'}) == 'bar'

    def test_extract_data_sacol(self):
        person = Person.testing_create()
        col = SheetColumn(Person.firstname)
        assert col.extract_data(person) == person.firstname

    def test_extract_data_sacol_altcolname(self):
        person = Person.testing_create(lastname=u'foo')
        col = SheetColumn(Person.lastname)
        assert col.extract_data(person) == 'foo'

    def test_extract_data_combined_sacol(self):
        person = Person.testing_create(intcol=8, floatcol=9)
        col = SheetColumn((Person.intcol, Person.floatcol, operator.add))
        assert col.extract_data(person) == 17

    def test_extract_data_combined_string(self):
        col = SheetColumn(('foo', 'bar', operator.add))
        assert col.extract_data({'foo': 9, 'bar': 8}) == 17

    def test_extract_data_combined_int(self):
        col = SheetColumn((9, 8, operator.add))
        assert col.extract_data({}) == 17

    def test_extract_data_combined_float(self):
        col = SheetColumn((9.5, 8, operator.add))
        assert col.extract_data({}) == 17.5

    def test_header_construction(self):
        class TestSheet(ReportSheet):
            def fetch_records(self):
                return []

            pre_data_rows = 5
            SheetColumn('key', header_2='foo', header_4='bar')

        sheet = TestSheet(Workbook(BytesIO()))
        col = sheet.units[0]
        assert col.header_data == ['', '', 'foo', '', 'bar']

    def test_header_construction_overflows_sheet(self):
        class TestSheet(ReportSheet):
            def fetch_records(self):
                return []

            pre_data_rows = 3
            SheetColumn('key', header_2='foo', header_4='bar')

        with pytest.raises(ProgrammingError) as exc_info:
            TestSheet(Workbook(BytesIO()))
        assert 'not enough pre-data rows on sheet' in str(exc_info)

    def test_width(self):
        col = SheetColumn('key')
        col.register_col_width('something')
        col.register_col_width('something else')
        assert col.xls_computed_width == 14


class TestLabeledColumn(object):
    def test_header_construction(self):
        col = LabeledColumn('foo\nbar', 'key')
        assert col._init_header == {0: 'foo', 1: 'bar'}

    def test_header_construction_offset(self):
        class _LabeledColumn(LabeledColumn):
            header_start_row = 4

        col = _LabeledColumn('foo\nbar', 'key')
        assert col._init_header == {4: 'foo', 5: 'bar'}


class TestOutput(object):
    @mock.patch('tribune.tests.reports.CarSheet.fetch_records', car_data)
    def generate_report(self, reportcls=CarSheet):
        wb = Workbook(BytesIO())
        ws = wb.add_worksheet(reportcls.sheet_name)
        reportcls(wb, worksheet=ws)
        book = xlsx_to_reader(wb)
        return book.sheet_by_name('Car Info')

    def test_header(self):
        ws = self.generate_report()
        assert ws.cell_value(0, 0) == 'Cars'

    def test_column_headings(self):
        def checkit(ws, row, column, value):
            assert ws.cell_value(row, column) == value

        ws = self.generate_report()
        for row, column, value in [
            (2, 0, ''),
            (2, 1, 'Year'),
            (2, 2, 'Make'),
            (2, 3, 'Model'),
            (2, 4, 'Style'),
            (2, 5, 'Color'),
            (2, 6, 'Blue'),
            (3, 6, 'Book'),
        ]:
            yield checkit, ws, row, column, value

    def test_column_data(self):
        def checkit(ws, row, column, value):
            assert ws.cell_value(row, column) == value

        ws = self.generate_report()
        for data_row, data in enumerate(car_data(None)):
            row = data_row + 4
            for column, value in [
                (0, ''),
                (find_sheet_col(ws, 'Year', 2), data['year']),
                (find_sheet_col(ws, 'Make', 2), data['make']),
                (find_sheet_col(ws, 'Model', 2), data['model']),
                (find_sheet_col(ws, 'Style', 2), data['style']),
                (find_sheet_col(ws, 'Color', 2), data['color']),
                (find_sheet_col(ws, 'Blue', 2), data['book_value']),
            ]:
                yield checkit, ws, row, column, value

    def test_merged_section_heading(self):
        class TestSheet(ReportSheet):
            class CarMergedHeaderSection(SheetSection):
                LabeledColumn('Style', 'style')
                LabeledColumn('Color', 'color')

                def write_header(self):
                    record_dict = {'foo': 'bar', 'baz': 'data'}
                    self.sheet.write_simple_merge(2, (record_dict, 'baz'))

            pre_data_rows = 1
            sheet_name = 'Car Info'
            CarMergedHeaderSection()

            def fetch_records(self):
                return []

        ws = self.generate_report(reportcls=TestSheet)
        assert ws.cell_value(0, 0) == 'data'

    def test_merged_section_heading_single_length(self):
        class TestSheet(ReportSheet):
            class CarMergedHeaderSection(SheetSection):
                LabeledColumn('Style', 'style')

                def write_header(self):
                    record_dict = {'foo': 'bar', 'baz': 'data'}
                    self.sheet.write_simple_merge(1, (record_dict, 'baz'))

            pre_data_rows = 1
            sheet_name = 'Car Info'
            CarMergedHeaderSection()

            def fetch_records(self):
                return []

        ws = self.generate_report(reportcls=TestSheet)
        assert ws.cell_value(0, 0) == 'data'


class TestPortraitOutput(object):
    @mock.patch('tribune.tests.reports.CarDealerSheet.fetch_records', dealer_data)
    def generate_report(self):
        wb = Workbook(BytesIO())
        ws = wb.add_worksheet(CarDealerSheet.sheet_name)
        CarDealerSheet(wb, worksheet=ws)
        book = xlsx_to_reader(wb)
        return book.sheet_by_name('Dealership Summary')

    def test_header(self):
        ws = self.generate_report()
        assert ws.cell_value(0, 0) == 'Dealership'

    def test_column_headings(self):
        def checkit(ws, row, column, value):
            assert ws.cell_value(row, column) == value

        ws = self.generate_report()
        for row, column, value in [
            (3, 0, ''),
            (3, 1, 'Count'),
            (3, 2, 'Income'),
            (3, 3, 'Expense'),
            (3, 4, 'Net'),
        ]:
            yield checkit, ws, row, column, value

    def test_row_data(self):
        def checkit(ws, row, column, value):
            assert ws.cell_value(row, column) == value

        ws = self.generate_report()
        row = find_sheet_row(ws, 'Sales', 0)
        for column, value in [
            (1, 5),
            (2, 75000),
            (3, 25000),
            (4, 50000),
        ]:
            yield checkit, ws, row, column, value

        row = find_sheet_row(ws, 'Rentals', 0)
        for column, value in [
            (1, 46),
            (2, 8600),
            (3, 2900),
            (4, 5700),
        ]:
            yield checkit, ws, row, column, value

        row = find_sheet_row(ws, 'Leases', 0)
        for column, value in [
            (1, 27),
            (2, 10548),
            (3, 3680),
            (4, 6868),
        ]:
            yield checkit, ws, row, column, value

        row = find_sheet_row(ws, 'Totals', 0)
        for column, value in [
            (4, 100000),
        ]:
            yield checkit, ws, row, column, value
